/*
 *	binkleyforce -- unix FTN mailer project
 *	
 *	Copyright (c) 1998-2000 Alexander Belkin, 2:5020/1398.11
 *	
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *	
 *	$Id: outb_fsqueue.c,v 1.2 2000/05/12 21:32:34 adb Exp $
 */

#include "includes.h"
#include "confread.h"
#include "logger.h"
#include "util.h"
#include "outbound.h"

int out_filetype(const char *fname)
{
	const char *p_nam;
	const char *p_ext;
	int i;
	
	/*
	 *  Get only file name, w/o dir part
	 */
	if( (p_nam = strrchr(fname, DIRSEPCHR)) )
		p_nam++;
	else
		p_nam = fname;
	
	/*
	 *  Get file name extension
	 */
	if( (p_ext = strrchr(p_nam, '.')) == NULL )
		return TYPE_UNKNOWN;
	
	for( i = 0; outtab[i].ext; i++ )
		if( strcasemask(p_ext, outtab[i].ext) == 0 )
			return outtab[i].type;
	
	for( i = 0; exttab[i].ext; i++ )
		if( strcasemask(p_ext, exttab[i].ext) == 0 )
			return exttab[i].type;
	
	return TYPE_UNKNOWN;
}

/* ------------------------------------------------------------------------- */
/* Read FLO file line by line and call handler proc for each file in FLO     */
/* ------------------------------------------------------------------------- */
static int out_readflo(s_filelist **fslist, const char *floname, int flavor,
                       bool aso, int flodsc)
{
	FILE *fp;
	char s[BF_MAXPATH+1], *p;
	s_filelist **tmpl;
	struct stat st;
	int type, action;
    
	DEB((D_OUTBOUND, "out_readflo: opening flo file \"%s\"", floname));
	
	if( (fp = fopen(floname, "r")) == NULL )
		{ logerr("can't open flo \"%s\"", floname); return 1; }
    
	for( tmpl = fslist; *tmpl; tmpl = &(*tmpl)->next );

	while(fgets(s, sizeof(s), fp))
	{
		string_chomp(s);
		if( *s && *s != '~' )
		{
			if( *s == '-' || *s == '^' )
				{ action = ACTION_UNLINK; p = s+1; }
			else if( *s == '#' )
				{ action = ACTION_TRUNCATE; p = s+1; }
			else if( *s == '@' )
				{ action = ACTION_NOTHING; p = s+1; }
			else
				{ action = ACTION_NOTHING; p = s; }
			
			if( stat(p, &st) == 0 )
			{
				type  = out_filetype(p);
				type |= TYPE_FROMFLO;
				type |= aso ? TYPE_ASONAME : 0;

				(*tmpl) = (s_filelist*)xmalloc(sizeof(s_filelist));
				memset(*tmpl, '\0', sizeof(s_filelist));
				(*tmpl)->fname   = (char *)xstrcpy(p);
				(*tmpl)->size    = st.st_size;
				(*tmpl)->type    = type;
				(*tmpl)->flavor  = flavor;
				(*tmpl)->action  = action;
				(*tmpl)->status  = STATUS_WILLSEND;
				(*tmpl)->flodsc  = flodsc;
				tmpl = &(*tmpl)->next;
			}
			else
				logerr("can't stat file \"%s\" from flo \"%s\"", p, floname);
		}
	}
	fclose(fp);
	return 0;
}

/*
 *  Add new entry to the flo files table, return it's index number
 */
static int out_addfloentry(s_fsqueue *q, const char *floname, int flavor)
{
	q->flotab = (s_flofile*)xrealloc(q->flotab, sizeof(s_flofile)*(q->flonum+1));

	memset(&q->flotab[q->flonum], '\0', sizeof(s_flofile));
	q->flotab[q->flonum].fname  = xstrcpy(floname);
	q->flotab[q->flonum].flavor = flavor;
	q->flonum++;

	return q->flonum-1;
}

/* ------------------------------------------------------------------------- */
/* Handle each file $fi found in outbound. Usually it called by outbound     */
/* scanner.. Pointer to result information will be stored at dst             */
/* ------------------------------------------------------------------------- */
void out_handle_fsqueue(char *dst, s_faddr addr, const char *fname, bool aso)
{
	int i;
	const char *p_ext;
	s_fsqueue *q = (s_fsqueue*)dst;

	DEB((D_OUTBOUND, "out_handle_fsqueue: process file \"%s\"", fname));
	
	if( (p_ext = strrchr(fname, '.')) == NULL ) return;

	for( i = 0; outtab[i].ext; i++ )
	{
		if( strcasecmp(p_ext, outtab[i].ext) == 0 ) break;
	}
	if( outtab[i].ext == NULL ) return;
	
	if( outtab[i].type == TYPE_NETMAIL || outtab[i].type == TYPE_REQUEST )
	{
		struct stat st;
		if( stat(fname, &st) == 0 )
		{
			s_filelist **tmpl;
			for( tmpl = &q->fslist; *tmpl; tmpl = &(*tmpl)->next );
			(*tmpl) = (s_filelist*)xmalloc(sizeof(s_filelist));
			memset(*tmpl, '\0', sizeof(s_filelist));
			(*tmpl)->fname   = (char *)xstrcpy(fname);
			(*tmpl)->size    = st.st_size;
			(*tmpl)->type    = outtab[i].type | (aso ? TYPE_ASONAME : 0);
			(*tmpl)->flavor  = outtab[i].flavor;
			(*tmpl)->action  = outtab[i].action;
			(*tmpl)->status  = STATUS_WILLSEND;
			(*tmpl)->flodsc  = 0;
		}
		else
			logerr("can't stat file \"%s\"", fname);
	}
	else if( outtab[i].type == TYPE_FLOFILE )
	{
		out_readflo(&q->fslist, fname, outtab[i].flavor, aso,
		            out_addfloentry(q, fname, outtab[i].flavor));
	}

	return;
}

void deinit_filelist(s_filelist *filelist)
{
	struct filelist *ptrl, *next;
	
	for( ptrl = filelist; ptrl; ptrl = next )
	{
		next = ptrl->next;
		if( ptrl->fname ) free(ptrl->fname);
		free(ptrl);
	}
}

void deinit_fsqueue(s_fsqueue *q)
{
	int i;

	if( q->fslist )
		deinit_filelist(q->fslist);

	if( q->flotab )
	{
		for( i = 0; i < q->flonum; i++ )
			if( q->flotab[i].fname ) free(q->flotab[i].fname);
		free(q->flotab);
	}
	
	memset(q, '\0', sizeof(s_fsqueue));
}

#ifdef DEBUG
void log_filelist(const s_filelist *filelist)
{
	const s_filelist *ptrl;
	
	DEB((D_OUTBOUND, "log_filelist: BEGIN"));
	for( ptrl = filelist; ptrl; ptrl = ptrl->next )
	{
		DEB((D_OUTBOUND, "log_filelist: \tfile \"%s\", %ld byte(s), flodsc = %d, status: \"%s\"",
		     ptrl->fname, (long)ptrl->size, ptrl->flodsc,
		     ptrl->status == STATUS_WILLSEND ? "WILLSEND" :
		     ptrl->status == STATUS_SKIP     ? "SKIP"     :
		     ptrl->status == STATUS_SKIP     ? "SENDING"  :
		     ptrl->status == STATUS_SENT     ? "SENT"     : "?"));
	}
	DEB((D_OUTBOUND, "log_filelist: END"));
}

void log_fsqueue(const s_fsqueue *q)
{
	int i;
	
	DEB((D_OUTBOUND, "log_queue_long: BEGIN"));

	log_filelist(q->fslist);

	for( i = 0; i < q->flonum; i++ )
		DEB((D_OUTBOUND, "log_queue_long: FLO file \"%s\"", q->flotab[i].fname));

	DEB((D_OUTBOUND, "log_queue_long: END"));
}

#endif
