/*
 *	binkleyforce -- unix FTN mailer project
 *	
 *	Copyright (c) 1998-2000 Alexander Belkin, 2:5020/1398.11
 *	
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *	
 *	$Id: prot_binkp.c,v 1.20 2000/10/28 12:07:51 adb Exp $
 */

#include "includes.h"
#include "confread.h"
#include "logger.h"
#include "util.h"
#include "io.h"
#include "session.h"
#include "outbound.h"
#include "prot_common.h"
#include "prot_binkp.h"

typedef enum {
	/*
	 *  Pseudo message types, returned by binkp_recv()
	 */
	BPMSG_EXIT = -3,	/* Terminate session */
	BPMSG_NONE = -2,	/* Got nothing intresting */
	BPMSG_DATA = -1,	/* Got data block */
	/*
	 *  Real BinkP message types
	 */
	BPMSG_NUL = 0,		/* Site information */
	BPMSG_ADR,		/* List of addresses */
	BPMSG_PWD,		/* Session password */
	BPMSG_FILE,		/* File information */
	BPMSG_OK,		/* Password was acknowleged (data ignored) */
	BPMSG_EOB,		/* End Of Batch (data ignored) */
	BPMSG_GOT,		/* File received */
	BPMSG_ERR,		/* Misc errors */
	BPMSG_BSY,		/* All AKAs are busy */
	BPMSG_GET,		/* Get a file from offset */
	BPMSG_SKIP,		/* Skip a file (RECEIVE LATER) */
	BPMSG_MIN = BPMSG_NUL,	/* Minimal message type value */
	BPMSG_MAX = BPMSG_SKIP	/* Maximal message type value */
} e_bpmsg;

typedef struct {
	bool sent;		/* Sent message (queued) */
	e_bpmsg type;		/* Message type */
	int size;		/* Message size (length of text) */
	char *data;		/* Message text */
} s_bpmsg;

typedef struct {
	char   *obuf;		/* Output buffer */
	int     opos;		/* Unsent bytes left in buffer */
	char   *optr;		/* Send data from this position */
	
	char   *ibuf;		/* Our input buffer */
	int     ipos;
	int     isize;
	bool    imsg;		/* Is it message? */
	bool    inew;
	e_bpmsg imsgtype;	/* Message type */

	int     junkcount;
	s_bpmsg *msgqueue;	/* Outgoing messages queue */
	int     n_msgs;		/* Number of messages in queue */

	int     timeout;
} s_bpinfo;

static void binkp_init(s_bpinfo *bpi);
static void binkp_deinit(s_bpinfo *bpi);
static int  binkp_gethdr(const char *s);
static int  binkp_getmsgtype(char ch);
static void binkp_puthdr(char *s, int u);
static void binkp_putmsgtype(char *s, e_bpmsg msg);
static void binkp_queuemsg(s_bpinfo *bpi, e_bpmsg msg, const char *s1, const char *s2);
static void binkp_queuemsgf(s_bpinfo *bpi, e_bpmsg msg, const char *fmt, ...);
static int  binkp_parsfinfo(char *s, char **fn, size_t *sz, time_t *tm, size_t *offs);
static int  binkp_buffer_message(s_bpinfo *bpi, s_bpmsg *msg);
static int  binkp_send_buffer(s_bpinfo *bpi);
static int  binkp_send(s_bpinfo *bpi);
static int  binkp_recv(s_bpinfo *bpi);
static void binkp_update_sysinfo(s_binkp *binkp);
static void binkp_log_sysinfo(s_binkp *binkp);
static void binkp_proc_NUL(e_bpmsg msg, char *s, s_binkp *binkp);
static void binkp_proc_ADR(e_bpmsg msg, char *s, s_binkp *binkp);
static void binkp_queue_sysinfo(s_bpinfo *bpi, s_binkp *binkp);

static void binkp_init(s_bpinfo *bpi)
{
	memset(bpi, '\0', sizeof(s_bpinfo));

	bpi->obuf  = (char*)xmalloc(BINKP_MAX_BLKSIZE + BINKP_BLK_HDRSIZE + 1);
	bpi->opos  = 0;
	bpi->optr  = bpi->obuf;
	
	bpi->inew  = TRUE;
	bpi->isize = -1;
	bpi->ibuf  = (char*)xmalloc(BINKP_MAX_BLKSIZE + BINKP_BLK_HDRSIZE + 1);

	bpi->timeout = conf_number(cf_binkp_timeout);
	if( !bpi->timeout )
		bpi->timeout = 300;
}

static void binkp_deinit(s_bpinfo *bpi)
{
	if( bpi->obuf )
		free(bpi->obuf);
	if( bpi->ibuf )
		free(bpi->ibuf);
	if( bpi->msgqueue )
		free(bpi->msgqueue);
	
	memset(bpi, '\0', sizeof(s_bpinfo));
}

int binkp_handshake(void)
{
	int  rc = 0;
	int  recv_rc = 0;
	int  send_rc = 0;
	bool send_ready = FALSE;
	bool recv_ready = FALSE;
	bool good_passw = FALSE;
	s_bpinfo bpi;

	session_set_freqs_status();

	binkp_init(&bpi);
	binkp_update_sysinfo(&state.tx.binkp);
	binkp_queue_sysinfo(&bpi, &state.tx.binkp);

	while(1)
	{
		send_ready = recv_ready = FALSE;
		
		if( tty_select(&recv_ready, (bpi.opos || bpi.n_msgs) ?
		               &send_ready : NULL, bpi.timeout) < 0 )
			gotoexit(PRC_ERROR);
		
		recv_rc = BPMSG_NONE;
		send_rc = 0;
		
		if( recv_ready && (recv_rc = binkp_recv(&bpi)) == BPMSG_EXIT )
			gotoexit(PRC_ERROR);
		
		if( send_ready && (send_rc = binkp_send(&bpi)) < 0 )
			gotoexit(PRC_ERROR);
	
		switch(recv_rc) {
		case BPMSG_NONE:
			break;
			
		case BPMSG_NUL:
			binkp_proc_NUL(recv_rc, bpi.ibuf+1, &state.rx.binkp);
			break;
			
		case BPMSG_ADR:
			binkp_proc_ADR(recv_rc, bpi.ibuf+1, &state.rx.binkp);
			if( state.caller )
			{
				if( session_addrs_check_genuine(state.rx.binkp.addrs,
							state.rx.binkp.anum, state.node.addr) )
				{
					state.error = SERR_NOADDRESS;
					goto Finish;
				}
				else
				{
					session_remote_lookup(state.rx.binkp.addrs,
							state.rx.binkp.anum);
				}
			}
			break;
			
		case BPMSG_PWD: /* Session password */
			if( !state.caller )
			{
				if( !strcmp(bpi.ibuf+1, "-") )
					state.rx.binkp.passwd[0] = '\0';
				else
					strnxcpy(state.rx.binkp.passwd, bpi.ibuf+1, sizeof(state.rx.binkp.passwd));
				
				if( !state.rx.binkp.anum )
				{
					binkp_queuemsg(&bpi, BPMSG_BSY, NULL, "No parsable addresses found");
					state.error = SERR_BUSY;
				}
				else if( session_addrs_check(state.rx.binkp.addrs,
				                        state.rx.binkp.anum,
				                        state.rx.binkp.passwd) )
				{
					binkp_queuemsg(&bpi, BPMSG_ERR, NULL, "Bad password");
					state.error = SERR_BADPASSWD;
				}
				else if( session_addrs_lock(state.rx.binkp.addrs,
				                            state.rx.binkp.anum) )
				{
					binkp_queuemsg(&bpi, BPMSG_BSY, NULL, "All AKAs are busy");
					state.error = SERR_BUSY;
				}
				else
				{
					binkp_queuemsg(&bpi, BPMSG_OK, NULL, NULL);
					good_passw = TRUE;
				}
				
				binkp_log_sysinfo(&state.rx.binkp);
				session_remote_log_status();
				
				goto Finish;
			}
			break;
			
		case BPMSG_OK: /* Password acknowledge */
			if( state.caller )
			{
				if( session_addrs_check(state.rx.binkp.addrs,
				                        state.rx.binkp.anum,
				                        state.tx.binkp.passwd) )
				{
					binkp_queuemsg(&bpi, BPMSG_ERR, NULL, "Bad password");
					state.error = SERR_BADPASSWD;
				}
				else
				{
					(void)session_addrs_lock(state.rx.binkp.addrs,
					                         state.rx.binkp.anum);
					good_passw = TRUE;
				}
				
				binkp_log_sysinfo(&state.rx.binkp);
				session_remote_log_status();
				
				goto Finish;
			}
			break;
			
		case BPMSG_ERR:
			log("remote report error: \"%s\"", bpi.ibuf+1);
			gotoexit(PRC_ERROR);
			break;
			
		case BPMSG_BSY:
			log("remote busy error: \"%s\"", bpi.ibuf+1);
			gotoexit(PRC_ERROR);
			break;
		}
	}

Finish:
	while( bpi.opos > 0 || bpi.n_msgs > 0 )
	{
		if( tty_select(NULL, &send_ready, 30) < 0 )
			gotoexit(PRC_ERROR);
		
		if( !send_ready || binkp_send(&bpi) < 0 )
			gotoexit(PRC_ERROR);
	}


exit:
	binkp_deinit(&bpi);
	
	if( !good_passw || state.error != SERR_NOERROR )
		return -1;
	
	if( session_set_inbound() )
		return -1;
	
	session_set_send_options();
	
	/*
	 * Create mail/files queue
	 */
	session_create_files_queue(state.rx.binkp.addrs,
	                           state.rx.binkp.anum);
	
	return 0;
}

int binkp_transfer(s_protinfo *pi)
{
	int  i, n, rc = PRC_NOERROR;
	bool recv_ready = FALSE;
	bool send_ready = FALSE;
	bool sent_EOB = FALSE;
	bool rcvd_EOB = FALSE;
	bool send_file = FALSE;
	bool recv_file = FALSE;
	bool wait_got = FALSE;
	bool nofiles = FALSE;
	int  recv_rc = 0;
	int  send_rc = 0;
	char *fname = NULL;
	size_t fsize = 0;
	time_t ftime = 0;
	size_t foffs = 0;
	s_bpinfo bpi;
	
	binkp_init(&bpi);
		
	while( !sent_EOB || !rcvd_EOB )
	{
		if( !send_file && !sent_EOB && bpi.opos == 0 && bpi.n_msgs == 0 )
		{
			if( !p_tx_fopen(pi) )
			{
				send_file = TRUE;
				binkp_queuemsgf(&bpi, BPMSG_FILE, "%s %ld %ld 0",
					pi->send->net_name, (long)pi->send->bytes_total,
					(long)pi->send->mod_time);
			}
			else
				/* No more files */
				nofiles = TRUE;
		}
		
		if( send_file && bpi.opos == 0 && bpi.n_msgs == 0 )
		{
			if( (n = p_tx_readfile(bpi.obuf + BINKP_BLK_HDRSIZE,
			                       4096, pi)) < 0 )
			{
				p_tx_fclose(pi);
				send_file = FALSE;
			}
			else
			{
				binkp_puthdr(bpi.obuf, (unsigned)(n & 0x7fff));
				bpi.opos = n + BINKP_BLK_HDRSIZE;
				pi->send->bytes_sent += n;
				if( pi->send->eofseen )
				{
					wait_got = TRUE; send_file = FALSE;
					pi->send->status = FSTAT_WAITACK;
				}
			}
		}

		if( nofiles && !wait_got && !sent_EOB )
		{
			sent_EOB = TRUE;
			binkp_queuemsg(&bpi, BPMSG_EOB, NULL, NULL);
		}

		recv_ready = send_ready = FALSE;
		if( tty_select(&recv_ready, (bpi.opos || bpi.n_msgs) ?
		               &send_ready : NULL, bpi.timeout) < 0 )
			gotoexit(PRC_ERROR);
		
		recv_rc = BPMSG_NONE;
		send_rc = 0;
		
		if( recv_ready && (recv_rc = binkp_recv(&bpi)) == BPMSG_EXIT )
			gotoexit(PRC_ERROR);
		if( send_ready && (send_rc = binkp_send(&bpi)) < 0 )
			gotoexit(PRC_ERROR);
	
		switch(recv_rc) {
		case BPMSG_NONE:
			break;
		case BPMSG_DATA: /* Got new data block */
			if( recv_file )
			{
				pi->recv->bytes_received += bpi.isize;
				if( (n = p_rx_writefile(bpi.ibuf, bpi.isize, pi)) < 0 )
				{
					/* error writing file */
					if( n == -2 )
					{
						binkp_queuemsgf(&bpi, BPMSG_GOT, "%s %ld %ld",
							pi->recv->net_name, (long)pi->recv->bytes_total,
							(long)pi->recv->mod_time);
					}
					else
					{
						binkp_queuemsgf(&bpi, BPMSG_SKIP, "%s %ld %ld",
							pi->recv->net_name, (long)pi->recv->bytes_total,
							(long)pi->recv->mod_time);
					}
					recv_file = FALSE;
					p_rx_fclose(pi);
				}
				else if( pi->recv->bytes_received == pi->recv->bytes_total )
				{
					recv_file = FALSE;
					pi->recv->status = FSTAT_SUCCESS;
					if( p_rx_fclose(pi) == 0 )
					{
						binkp_queuemsgf(&bpi, BPMSG_GOT, "%s %ld %ld",
							pi->recv->net_name, (long)pi->recv->bytes_total,
							(long)pi->recv->mod_time);
					}
					else
					{
						binkp_queuemsgf(&bpi, BPMSG_SKIP, "%s %ld %ld",
							pi->recv->net_name, (long)pi->recv->bytes_total,
							(long)pi->recv->mod_time);
					}
				}
			}
#ifdef DEBUG
			else DEB((D_PROT, "ignore received data block"));
#endif
			break;
			
		case BPMSG_FILE:
			if( recv_file )
				{ p_rx_fclose(pi); recv_file = FALSE; }
			
			if( binkp_parsfinfo(bpi.ibuf+1, &fname,
			    &fsize, &ftime, &foffs) )
			{
				binkp_queuemsg(&bpi, BPMSG_ERR, "FILE: ", "unparsable arguments");
				goto FinishSession;
			}
			
			if( pi->recv && !p_compfinfo(pi->recv, fname, fsize, ftime)
			 && pi->recv->bytes_skipped == foffs )
			{
				recv_file = TRUE;
				break;
			}
			
			switch(p_rx_fopen(pi, fname, fsize, ftime, 0)) {
			case 0:
				if( pi->recv->bytes_skipped == 0 )
				{
					recv_file = TRUE;
					break;
				}
				binkp_queuemsgf(&bpi, BPMSG_GET, "%s %ld %ld %ld",
					pi->recv->net_name, (long)pi->recv->bytes_total,
					(long)pi->recv->mod_time,
					(long)pi->recv->bytes_skipped);
				break;
			case 1:
				/* SKIP (non-destructive) */
				binkp_queuemsgf(&bpi, BPMSG_SKIP, "%s %ld %ld",
					pi->recv->net_name, (long)pi->recv->bytes_total,
					(long)pi->recv->mod_time);
				break;
			case 2:
				/* SKIP (destructive) */
				binkp_queuemsgf(&bpi, BPMSG_GOT, "%s %ld %ld",
					pi->recv->net_name, (long)pi->recv->bytes_total,
					(long)pi->recv->mod_time);
				break;
			default:
				ASSERT_MSG();
			}
			break;
			
		case BPMSG_EOB: /* End Of Batch */
			if( recv_file )
			{
				p_rx_fclose(pi);
				recv_file = FALSE;
			}
			rcvd_EOB = TRUE;
			break;
			
		case BPMSG_GOT:
		case BPMSG_SKIP:
			if( binkp_parsfinfo(bpi.ibuf+1, &fname, &fsize, &ftime, NULL) == 0 )
			{
				if( send_file && !p_compfinfo(pi->send, fname, fsize, ftime) )
				{
					/* We got GOT/SKIP for current file! */
					if( recv_rc == BPMSG_GOT )
						pi->send->status = FSTAT_SKIPPED;
					else
						pi->send->status = FSTAT_REFUSED;
					p_tx_fclose(pi);
					send_file = FALSE;
					break;
				}
				
				wait_got = FALSE;
				
				for( i = 0; i < pi->n_sentfiles; i++ )
				{
					if( pi->sentfiles[i].status == FSTAT_WAITACK
					 && !p_compfinfo(&pi->sentfiles[i], fname, fsize, ftime) )
					{
						s_finfo *tmp = pi->send;
						
						pi->send = &pi->sentfiles[i];
						if( recv_rc == BPMSG_GOT )
							pi->send->status = FSTAT_SUCCESS;
						else
							pi->send->status = FSTAT_REFUSED;
						p_tx_fclose(pi);
						
						pi->send = tmp;
						break;
					}
					else if( pi->sentfiles[i].status == FSTAT_WAITACK )
						wait_got = TRUE;
				}

				/*
				 * Check, maybe there are files still
				 * waiting for the GOT/SKIP acknwoledge
				 */
				while( i < pi->n_sentfiles && !wait_got )
				{
					if( pi->sentfiles[i].status == FSTAT_WAITACK )
						wait_got = TRUE;
					++i;
				}
			}
			break;
			
		case BPMSG_ERR:
			log("remote report error: \"%s\"", bpi.ibuf+1);
			gotoexit(PRC_ERROR);
			break;
			
		case BPMSG_BSY:
			log("remote busy error: \"%s\"", bpi.ibuf+1);
			gotoexit(PRC_ERROR);
			break;
			
		case BPMSG_GET:
			if( binkp_parsfinfo(bpi.ibuf+1, &fname, &fsize, &ftime, &foffs) == 0 )
			{
				if( send_file && !p_compfinfo(pi->send, fname, fsize, ftime) )
				{
					if( fseek(pi->send->fp, foffs, SEEK_SET) == -1 )
					{
						log("cannot send file from requested offset %ld", (long)foffs);
						p_tx_fclose(pi);
						send_file = FALSE;
					}
					else
					{
						log("sending \"%s\" from %ld offset",
							pi->send->fname, (long)foffs);
						pi->send->bytes_skipped = foffs;
						pi->send->bytes_sent = foffs;
						binkp_queuemsgf(&bpi, BPMSG_FILE, "%s %ld %ld %ld",
							pi->send->net_name, (long)pi->send->bytes_total,
							(long)pi->send->mod_time, (long)foffs);
					}
				}
			}
			break;
			
		default:
			log("binkp got unhandled msg #%d", recv_rc);
			break;
		}
	} /* end of while( !sent_EOB || !rcvd_EOB ) */
	
FinishSession:
	while( bpi.opos > 0 || bpi.n_msgs > 0 )
	{
		if( tty_select(NULL, &send_ready, 30) < 0 )
			gotoexit(PRC_ERROR);
		
		if( !send_ready || binkp_send(&bpi) < 0 )
			gotoexit(PRC_ERROR);
	}
	FLUSHOUT();
	
exit:
	if( pi->send && pi->send->fp ) p_tx_fclose(pi);
	if( pi->recv && pi->recv->fp ) p_rx_fclose(pi);

	binkp_deinit(&bpi);
	
	DEB((D_PROT, "binkp: BINKP exit = %d", rc));
	
	return rc;
}

/* ------------------------------------------------------------------------- */
/* Fills s[0] and s[1] with binkp frame header using value of u              */
/* ------------------------------------------------------------------------- */
static void binkp_puthdr(char *s, int val)
{
	s[0] = ( (unsigned long) val >> 8 ) & 0xff;
	s[1] = ( (unsigned long) val      ) & 0xff;
}

/* ------------------------------------------------------------------------- */
/* Fills s[0] with binkp message type value                                  */
/* ------------------------------------------------------------------------- */
static void binkp_putmsgtype(char *s, int msg)
{
	s[0] = ( (unsigned int) msg ) & 0xff;
}

/* ------------------------------------------------------------------------- */
/* Extract size from binkp frame header                                      */
/* ------------------------------------------------------------------------- */
static int binkp_gethdr(const char *s)
{
	return ( (unsigned int) (((unsigned char) s[0]) & 0x7f) << 8 )
	     | ( (unsigned int) (((unsigned char) s[1]) & 0xff)      );
}

/* ------------------------------------------------------------------------- */
/* Extract message type from ch                                              */
/* ------------------------------------------------------------------------- */
static int binkp_getmsgtype(char ch)
{
	int val = ( ((unsigned char)ch) & 0x7f );
	
	if( BPMSG_MIN > val || val > BPMSG_MAX )
		return -1;
	else
		return val;
}

/* ------------------------------------------------------------------------- */
/* Puts a message to the output msg. queue. These msgs will be send          */
/* right after the current data block.                                       */
/* ------------------------------------------------------------------------- */
static void binkp_queuemsg(s_bpinfo *bpi, e_bpmsg msg, const char *s1,
                           const char *s2)
{
	bpi->msgqueue = xrealloc(bpi->msgqueue, sizeof(s_bpmsg)*(bpi->n_msgs + 1));
	memset(&bpi->msgqueue[bpi->n_msgs], '\0', sizeof(s_bpmsg));
	
	/* Set message type */
	bpi->msgqueue[bpi->n_msgs].type = msg;
	
	/* Set message data size (without frame header) */
	bpi->msgqueue[bpi->n_msgs].size = 1;
	if( s1 ) bpi->msgqueue[bpi->n_msgs].size += strlen(s1);
	if( s2 ) bpi->msgqueue[bpi->n_msgs].size += strlen(s2);
	
	/* Set message data */
	bpi->msgqueue[bpi->n_msgs].data = xmalloc(bpi->msgqueue[bpi->n_msgs].size+3);
	binkp_puthdr(bpi->msgqueue[bpi->n_msgs].data,
	             (unsigned)(bpi->msgqueue[bpi->n_msgs].size | 0x8000));
	
	binkp_putmsgtype(bpi->msgqueue[bpi->n_msgs].data + 2, msg);
	
	bpi->msgqueue[bpi->n_msgs].data[3] = '\0';
	if( s1 ) strcat(bpi->msgqueue[bpi->n_msgs].data + 3, s1);
	if( s2 ) strcat(bpi->msgqueue[bpi->n_msgs].data + 3, s2);

	DEB((D_PROT, "binkp_queuemsg: queued message #%d, %ld bytes, \"%s\"",
		(int)bpi->msgqueue[bpi->n_msgs].type,
		(long)bpi->msgqueue[bpi->n_msgs].size,
		(char*)bpi->msgqueue[bpi->n_msgs].data+3));

	++bpi->n_msgs;
}

/* ------------------------------------------------------------------------- */
/* Sends a message using format string                                       */
/* ------------------------------------------------------------------------- */
static void binkp_queuemsgf(s_bpinfo *bpi, e_bpmsg msg, const char *fmt, ...)
{
	char msg_text[2048];
	va_list ap;
	
	va_start(ap, fmt);
#ifdef HAVE_SNPRINTF
	vsnprintf(msg_text, sizeof(msg_text), fmt, ap);
#else
	vsprintf(msg_text, fmt, ap);
#endif
	va_end(ap);
	
	binkp_queuemsg(bpi, msg, msg_text, NULL);
}

/* ------------------------------------------------------------------------- */
/* Parse most common message arguments, send error message to remote         */
/* if needed. Return non-zero value on error.                                */
/* ------------------------------------------------------------------------- */
static int binkp_parsfinfo(char *s, char **fn, size_t *sz, time_t *tm,
                           size_t *offs)
{
	char *n;
	char *p_fname = NULL;
	char *p_size  = NULL;
	char *p_time  = NULL;
	char *p_offs  = NULL;
	
	/* Attention, offs may be NULL! */
	
	ASSERT(s != NULL && fn != NULL && sz != NULL && tm != NULL);

	DEB((D_PROT, "binkp_parsemsg: want parse \"%s\"", s));
	
	p_fname = string_token(s, &n, NULL, 0);
	p_size  = string_token(NULL, &n, NULL, 0);
	p_time  = string_token(NULL, &n, NULL, 0);
	if( offs ) p_offs = string_token(NULL, &n, NULL, 0);
	
	if( p_fname && p_size && p_time && (!offs || p_offs) )
	{
		if( ISDEC(p_size) && ISDEC(p_time) && (!offs || ISDEC(p_offs)) )
		{
			(*fn) = p_fname;
			(*sz) = atol(p_size);
			(*tm) = atol(p_time);
			if( offs ) *offs = atol(p_offs);
			DEB((D_PROT, "binkp_parsemsg: file = \"%s\", size = %d, time = %d",
				*fn, *sz, *tm));
			return 0;
		}
	}
	return 1;
}

/* ------------------------------------------------------------------------- */
/* Buffer new message. Return non-zero if can't put more messages in buffer  */
/* ------------------------------------------------------------------------- */
static int binkp_buffer_message(s_bpinfo *bpi, s_bpmsg *msg)
{
	DEB((D_PROT, "binkp_buffer_message: buffer message #%d, %ld byte(s)",
		(int)msg->type, (long)msg->size));
	
	/* Check for possible internal error */
	if( msg->size > BINKP_MAX_BLKSIZE )
	{
		log("size of msg we want to send is too big (%db)", msg->size);
		return 1;
	}
	
	/* Is there space for the new msg? */
	if( bpi->opos + msg->size > BINKP_MAX_BLKSIZE ) return 1;
	
	if( msg->data )
	{
		memcpy(bpi->obuf + bpi->opos, msg->data, msg->size + BINKP_BLK_HDRSIZE);
		bpi->opos += msg->size + BINKP_BLK_HDRSIZE;
		free(msg->data);
		msg->data = NULL;
	}
	
	return 0;
}

/* ------------------------------------------------------------------------- */
/* Try to send buffered data. Return zero if no errors happened              */
/* ------------------------------------------------------------------------- */
static int binkp_send_buffer(s_bpinfo *bpi)
{
	int n;

	DEB((D_PROT, "binkp_send_buffer: sending %ld byte(s)", (long)bpi->opos));
	
	n = tty_write(bpi->optr, bpi->opos);
	
	if( n >= bpi->opos )
		{ bpi->optr = bpi->obuf; bpi->opos = 0; }
	else if( n > 0 )
		{ bpi->optr += n; bpi->opos -= n; }
	else
		return -1;
	
	return 0;
}

/* ------------------------------------------------------------------------- */
/* Sends buffered data or msgs from the msg queue.                           */
/* ------------------------------------------------------------------------- */
static int binkp_send(s_bpinfo *bpi)
{
	int i;
	
	if( bpi->opos == 0 && bpi->msgqueue )
	{
		/*
		 *  Buffer is empty and there are unsent msgs
		 */
		for( i = 0; i < bpi->n_msgs; i++ )
		{
			if( bpi->msgqueue[i].sent == FALSE )
			{
				if( binkp_buffer_message(bpi, &bpi->msgqueue[i]) )
					break;
				else
					bpi->msgqueue[i].sent = TRUE;
			}
		}
		/* If the message queue is empty, free it */
		if( i >= bpi->n_msgs )
		{
			if( bpi->msgqueue )
				{ free(bpi->msgqueue); bpi->msgqueue = NULL; }
			bpi->n_msgs = 0;
		}
	}

	return bpi->opos ? binkp_send_buffer(bpi) : 0;
}

static int binkp_recv(s_bpinfo *bpi)
{
	int n = 0;
	int size;

	if( bpi->inew || bpi->isize == -1 )
	{
		bpi->inew  = FALSE;
		bpi->isize = -1;
		size = BINKP_BLK_HDRSIZE;
	}
	else
	{
		size = bpi->isize;
	}
	
	if( size > 0 )
	{
		n = tty_read(bpi->ibuf + bpi->ipos, size - bpi->ipos);
		if( n <= 0 ) return BPMSG_EXIT;
	}
	
	bpi->ipos += n;
	
	if( bpi->ipos == size )
	{
		if( bpi->isize == -1 )
		{
			/* We got block header */
			bpi->ipos  = 0;
			bpi->imsg  = ((bpi->ibuf[0] >> 7) & 0377) ? TRUE : FALSE;
			bpi->isize = binkp_gethdr(bpi->ibuf);
			DEB((D_PROT, "binkp_recv: received header: %ld byte(s) (%s)",
				(long)bpi->isize, bpi->imsg ? "msg" : "data"));
			if( bpi->isize > BINKP_MAX_BLKSIZE )
			{
				log("internal error: got %ld bytes block size", bpi->isize);
				return BPMSG_EXIT;
			}
			else if( bpi->isize == 0 )
			{
				bpi->ipos  = 0;
				bpi->inew  = TRUE;
				bpi->isize = -1;
				if( bpi->imsg == TRUE )
					log("zero length message from remote");
			}
		}
		else /* We got whole data block/message */
		{
			if( bpi->imsg == TRUE ) /* Is it message? */
			{
				bpi->ipos = 0;
				bpi->inew = TRUE;
				bpi->ibuf[size] = '\0';
				bpi->imsgtype = binkp_getmsgtype(bpi->ibuf[0]);
				DEB((D_PROT, "binkp_recv: got message #%d, %ld byte(s), \"%s\"",
					(int)bpi->imsgtype, bpi->isize, bpi->ibuf+1));
				if( bpi->imsgtype < 0 )
				{
					bpi->isize = -1;
					log("got incorrect message type");
					if( ++bpi->junkcount >= 5 )
					{
						log("junk count exceeded");
						return BPMSG_EXIT;
					}
					return BPMSG_NONE;
				}
				return bpi->imsgtype;
			}
			else /* It is data */
			{
				bpi->ipos = 0;
				bpi->inew = TRUE;
				bpi->ibuf[size] = '\0';
				DEB((D_PROT, "binkp_recv: got data block, %ld byte(s)",
					(long)bpi->isize));
				return BPMSG_DATA;
			}
		}
	}
	
	return BPMSG_NONE;
}

static void binkp_queue_sysinfo(s_bpinfo *bpi, s_binkp *binkp)
{
	int i;
	char *astr = NULL;
	char abuf[BF_MAXADDRSTR+1];
	
	binkp_queuemsg(bpi, BPMSG_NUL, "SYS ", binkp->systname);
	binkp_queuemsg(bpi, BPMSG_NUL, "ZYZ ", binkp->sysop);
	binkp_queuemsg(bpi, BPMSG_NUL, "LOC ", binkp->location);
	binkp_queuemsg(bpi, BPMSG_NUL, "PHN ", binkp->phone);
	binkp_queuemsg(bpi, BPMSG_NUL, "NDL ", binkp->flags);

	binkp_queuemsg(bpi, BPMSG_NUL, "TIME ", binkp->timestr);
	
	binkp_queuemsgf(bpi, BPMSG_NUL, "VER %s %s/%d.%d",
		binkp->progname, binkp->protname,
		binkp->majorver, binkp->minorver);
	
	for( i = 0; i < binkp->anum; i++ )
	{
		if( i ) astr = xstrcat(astr, " ");
		astr = xstrcat(astr, ftn_addrstr(abuf, binkp->addrs[i].addr));
	}
	
	if( astr )
		{ binkp_queuemsg(bpi, BPMSG_ADR, NULL, astr); free(astr); }
	
	if( state.caller )
	{
		if( *binkp->passwd )
			binkp_queuemsg(bpi, BPMSG_PWD, NULL, binkp->passwd);
		else
			binkp_queuemsg(bpi, BPMSG_PWD, NULL, "-");
	}
}

static void binkp_log_sysinfo(s_binkp *binkp)
{
	int i;
	char abuf[BF_MAXADDRSTR+1];
	
	if( binkp->anum )
		for( i = 0; i < binkp->anum; i++ )
		{
			log("   Address : %s", ftn_addrstr(abuf, binkp->addrs[i].addr));
		}
	else
		log("   Address : <none>");
	
	if( binkp->systname[0] && binkp->phone[0] )
		log("    System : %s (%s)",
			string_printable(binkp->systname),
			string_printable(binkp->phone));
	else if( binkp->systname[0] )
		log("    System : %s",
			string_printable(binkp->systname));
	else if( binkp->phone[0] )
		log("     Phone : %s",
			string_printable(binkp->phone));

#ifdef BFORCE_LOG_PASSWD
	if( binkp->passwd[0] )
		log("  Password : %s",
			string_printable(binkp->passwd));
#endif

	if( binkp->sysop[0] && binkp->location[0] )
		log("     SysOp : %s from %s",
			string_printable(binkp->sysop),
			string_printable(binkp->location));
	else if( binkp->sysop[0] )
		log("     SysOp : %s",
			string_printable(binkp->sysop));
	else if( binkp->location[0] )
		log("  Location : %s",
			string_printable(binkp->location));

	if( binkp->progname[0] )
	{
		log("    Mailer : %s (%s/%d.%d)",
			binkp->progname[0] ? string_printable(binkp->progname) : "?",
			binkp->protname[0] ? string_printable(binkp->protname) : "?",
			binkp->majorver, binkp->minorver);
	}
	if( binkp->flags[0] )
		log("     Flags : %s",
			string_printable(binkp->flags));
	if( binkp->timestr[0] )
		log("      Time : %s",
			string_printable(binkp->timestr));
}

static void binkp_proc_NUL(e_bpmsg msg, char *s, s_binkp *binkp)
{
	char *p;
	
	if( strncmp(s, "SYS ", 4) == 0 )
		strnxcpy(binkp->systname, s+4, sizeof(binkp->systname));
	else if( strncmp(s, "ZYZ ", 4) == 0 )
		strnxcpy(binkp->sysop, s+4, sizeof(binkp->sysop));
	else if( strncmp(s, "LOC ", 4) == 0 )
		strnxcpy(binkp->location, s+4, sizeof(binkp->location));
	else if( strncmp(s, "PHN ", 4) == 0 )
		strnxcpy(binkp->phone, s+4, sizeof(binkp->phone));
	else if( strncmp(s, "NDL ", 4) == 0 )
		strnxcpy(binkp->flags, s+4, sizeof(binkp->flags));
	else if( strncmp(s, "TIME ", 5) == 0 )
		strnxcpy(binkp->timestr, s+5, sizeof(binkp->timestr));
	else if( strncmp(s, "OPT ", 4) == 0 )
	{
	}
	else if( strncmp(s, "VER ", 4) == 0 )
	{
		s += 4;
		if( (p = strchr(s, ' ')) )
		{
			strnxcpy(binkp->progname, s, MIN(sizeof(binkp->progname), p-s+1));
			s = p+1;
			if( (p = strchr(s, '/')) )
			{
				strnxcpy(binkp->protname, s, MIN(sizeof(binkp->protname), p-s+1));
				sscanf(p+1, "%d.%d", &binkp->majorver, &binkp->minorver);
			}
		}
		else
		{
			strnxcpy(binkp->progname, s+4, sizeof(binkp->progname));
		}
	}
	else
	{
		log("binkp NUL: \"%s\"", s);
	}
}

static void binkp_proc_ADR(e_bpmsg msg, char *s, s_binkp *binkp)
{
	char *p, *q;
		
	if( binkp->addrs )
	{
		free(binkp->addrs);
		binkp->addrs = NULL;
		binkp->anum = 0;
	}
	
	for( p = string_token(s, &q, NULL, 0); p;
	     p = string_token(NULL, &q, NULL, 0) )
	{
		s_faddr addr;

		if( ftn_addrparse(&addr, p, FALSE) )
			log("unparsable address \"%s\" skipped", p);
		else
			session_addrs_add(&binkp->addrs, &binkp->anum, addr);
	}
	
	if( !binkp->anum )
		log("warning: no valid addresses");
}

static void binkp_update_sysinfo(s_binkp *binkp)
{
	s_cval_entry *addr_ptr;
	s_cval_entry *hide_ptr;
	s_faddr *primary = NULL;
	
	const char *p_systname  = conf_string(cf_system_name);
	const char *p_location  = conf_string(cf_location);
	const char *p_sysop     = conf_string(cf_sysop_name);
	const char *p_phone     = conf_string(cf_phone);
	const char *p_flags     = conf_string(cf_flags);
	
	/* free previously allocated memory */
	if( binkp->addrs )
		free(binkp->addrs);
		
	memset(binkp, '\0', sizeof(s_binkp));
	
	/* Set best primary address */
	if( state.caller )
		primary = session_get_bestaka(state.node.addr);
	else if( state.rx.binkp.anum > 0 )
		primary = session_get_bestaka(state.rx.binkp.addrs[0].addr);
		
	/* Add primary address */
	if( primary )
		session_addrs_add(&binkp->addrs, &binkp->anum, *primary);

	/* Add other AKAs */
	for( addr_ptr = conf_first(cf_address); addr_ptr;
	     addr_ptr = conf_next(addr_ptr) )
	{
		for( hide_ptr = conf_first(cf_hide_our_aka); hide_ptr;
		     hide_ptr = conf_next(hide_ptr) )
		{
			if( !ftn_addrcomp(hide_ptr->d.falist.addr, addr_ptr->d.falist.addr) )
				break;
		}
		
		if( !hide_ptr && primary != &addr_ptr->d.falist.addr )
			session_addrs_add(&binkp->addrs, &binkp->anum,
					addr_ptr->d.falist.addr);
	}
	
	if( binkp->anum == 0 )
		log("warning: no addresses will be presented to remote");

	/* session password */
	if( state.caller )
		session_get_password(state.node.addr, binkp->passwd, sizeof(binkp->passwd));
	
	binkp->majorver = BINKP_MAJOR;
	binkp->minorver = BINKP_MINOR;
	
	strnxcpy(binkp->progname, BF_NAME"/"BF_VERSION"/"BF_OS, sizeof(binkp->progname));
	strnxcpy(binkp->protname, BINKP_NAME, sizeof(binkp->protname));
	
	strnxcpy(binkp->sysop,    p_sysop    ? p_sysop    : "Unknown", sizeof(binkp->sysop));
	strnxcpy(binkp->systname, p_systname ? p_systname : "Unknown", sizeof(binkp->systname));
	strnxcpy(binkp->location, p_location ? p_location : "Unknown", sizeof(binkp->location));
	strnxcpy(binkp->phone,    p_phone    ? p_phone    : "Unknown", sizeof(binkp->phone));
	strnxcpy(binkp->flags,    p_flags    ? p_flags    : "BINKP",   sizeof(binkp->flags));

	time_string_format(binkp->timestr, sizeof(binkp->timestr), "%Y/%m/%d %H:%M:%S", 0);
}

